from __future__ import annotations
from typing import Type

from namespacecraft.term import Term


class Namespace:
    __slots__ = ('_base', '_hash', '_term_class', '_trailing_delim', '_last_has_delim')

    def __init__(self, base: str, *, term_cls: Type = Term, _trailing_delim: str | None = None, _last_has_delim: bool = False) -> None:
        base = str(base)

        if base.count('#') > 1:
            raise ValueError('A namespace may contain at most one fragment delimiter')

        self._hash = base.endswith('#')

        # Remove embedded fragment; only keep base
        if '#' in base:
            base, _ = base.split('#', 1)

        self._base = base
        self._term_class = term_cls
        self._trailing_delim = _trailing_delim
        self._last_has_delim = _last_has_delim

    def __truediv__(self, other: str | int | list | tuple) -> Namespace:
        if isinstance(other, (list, tuple)):
            ns = self
            for part in other:
                ns = ns / part
            return ns

        other_str = str(other)

        if self._hash:
            if '/' in other_str:
                raise ValueError('Cannot append hierarchical path to a hash namespace')
            return self._term_class(f'{self._base}#{other_str}')

        if '#' in other_str:
            raise ValueError('Cannot append fragment-like string to a slash namespace')

        # Normalised path join — exactly one '/'
        base = self._base.rstrip('/')
        part = other_str.lstrip('/')
        new_base = f'{base}/{part}'

        return Namespace(
            new_base,
            term_cls=self._term_class,
            _trailing_delim=self._trailing_delim,
            _last_has_delim=other_str.endswith('/')
        )

    def __add__(self, other: str):
        """Create a Term by literal concatenation, respecting configured termination."""
        other_str = str(other)

        if self._hash:
            return self._term_class(f'{self._base}#{other_str}')

        base_str = str(self)

        # If namespace was configured to terminate with a delimiter,
        # insert it exactly once at the termination boundary.
        if self._trailing_delim:
            if not base_str.endswith(self._trailing_delim) and not other_str.startswith(self._trailing_delim):
                return self._term_class(base_str + self._trailing_delim + other_str)
            if base_str.endswith(self._trailing_delim) and other_str.startswith(self._trailing_delim):
                return self._term_class(base_str + other_str.lstrip(self._trailing_delim))

        return self._term_class(base_str + other_str)

    def __getattr__(self, name: str):
        """Create a Term() by accessing an attribute"""
        if name.startswith('_'):
            raise AttributeError(f'{type(self).__name__} object has no attribute {name!r}')

        if self._hash:
            term_str = f'{self._base}#{name}'
        else:
            term_str = f'{self._base.rstrip('/')}/{name}'

        return self._term_class(term_str)

    def __getitem__(self, key: str):
        """Create a Term() by accessing an attribute by name"""
        return self.__getattr__(str(key))

    def __str__(self) -> str:
        return self._base + ('#' if self._hash else '')

    def __repr__(self) -> str:
        return f'Namespace({str(self)!r})'

    def __contains__(self, other: str | Namespace) -> bool:
        return str(other).startswith(self._base)

    @property
    def uri(self):
        """Return this namespace as a terminal URI (_term_class instance)"""
        return self._term_class(str(self))

    def terminates_with(self, character: str) -> Namespace:
        """Return a Namespace() guaranteed to end with the given delimiter"""
        if not character:
            raise ValueError('character must be a non-empty string')

        if self._hash:
            raise ValueError('Cannot set a trailing delimiter on a hash namespace')

        if self._base.endswith(character):
            return self

        return Namespace(
            self._base + character,
            term_cls=self._term_class,
            _trailing_delim=character
        )



import pytest
from namespacecraft import Namespace, Term


def test_term_is_str_subclass():
    term = Term('http://example.org/a')
    assert isinstance(term, str)
    assert isinstance(term, Term)


def test_term_is_not_namespace():
    term = Term('http://example.org/a')
    assert not isinstance(term, Namespace)


def test_term_repr():
    term = Term('http://example.org/a')
    assert repr(term) == "Term('http://example.org/a')"


def test_term_is_terminal_no_path_extension():
    term = Term('http://example.org/a')

    with pytest.raises(TypeError):
        term / 'b'


def test_term_has_no_attribute_namespace_semantics():
    term = Term('http://example.org/a')

    # Attribute access should behave like str, not Namespace
    with pytest.raises(AttributeError):
        _ = term.b



import pytest

from namespacecraft.namespace import Namespace, Term


def test_as_str():
    EX = Namespace('http://example.org/')
    assert str(EX) == 'http://example.org/'


def test_coerce_int_to_str():
    EX = Namespace('http://example.org/')
    PART = EX / 'part'
    assert str(PART / 1) == 'http://example.org/part/1'


def test_create_term_by_addition():
    EX = Namespace('http://example.org/')
    assert EX / 'a/b/c/' + 'a' == 'http://example.org/a/b/c/a'
    assert EX / 'x/y' + 'z' == 'http://example.org/x/yz'


def test_create_term_by_addition_with_leading_delimiter():
    EX = Namespace('http://example.org')
    term = EX + '/a'
    assert str(term) == 'http://example.org/a'


def test_create_term_without_delimiter():
    EX = Namespace('http://example.org')
    term = EX + 'a'
    assert str(term) == 'http://example.orga'


def test_create_term_by_getting_item():
    EX = Namespace('http://example.org/')
    
    t = EX['abc']
    assert isinstance(t, Term)
    assert str(t) == 'http://example.org/abc'
    
    # Should also work with paths
    t2 = EX['x/y/z']
    assert isinstance(t2, Term)
    assert str(t2) == 'http://example.org/x/y/z'
    
    # Dot notation and indexing should produce the same result for simple names
    assert str(EX.a) == str(EX['a'])


def test_deduplicates_delimiter_in_path():
    EX = Namespace('http://example.org/') / '/a'
    assert str(EX) == 'http://example.org/a'


def test_paths_from_list():
    EX = Namespace('http://example.org/') / [1, 2, 3]
    assert str(EX) == 'http://example.org/1/2/3'


def test_paths_from_list_and_create_term():
    EX = Namespace('http://example.org/') / [1, 2, 3] + '/a'
    assert str(EX) == 'http://example.org/1/2/3/a'


def test_create_term_from_attribute():
    EX = Namespace('http://example.org/')
    a = EX.a
    assert str(a) == 'http://example.org/a'
    assert isinstance(a, Term)

    EX = Namespace('http://example.org#')
    b = EX.b
    assert str(b) == 'http://example.org#b'
    assert isinstance(b, Term)


def test_create_path_with_hash_delimiter():
    EX = Namespace('http://example.org#') / 'a'
    assert str(EX) == 'http://example.org#a'

# Expected errors

def test_only_one_fragment_allowed():
    EX = Namespace('http://example.org#')
    
    # First fragment is fine
    assert EX + 'a' == 'http://example.org#a'
    
    # The original Namespace can still be used to create another fragment
    assert EX + 'b' == 'http://example.org#b'


def test_mismatched_delimiters_in_path():
    # Slash namespace cannot accept #
    EX_slash = Namespace('http://example.org/')
    with pytest.raises(ValueError):
        EX_slash / '#a'

    # Hash namespace cannot accept /
    EX_hash = Namespace('http://example.org#')
    with pytest.raises(ValueError):
        EX_hash / '/a'


def test_hash_delimiter_is_terminal():
    EX = Namespace('http://example.org#')

    # First / returns a terminal object
    term = EX / 'a'
    assert term == 'http://example.org#a'

    # Further / calls are invalid (TypeError)
    with pytest.raises(TypeError):
        term / 'b'


def test_add_returns_term():
    EX = Namespace('http://example.org/')
    term = EX + 'a'

    assert isinstance(term, Term)
    assert term == 'http://example.org/a'


def test_dot_returns_term():
    EX = Namespace('http://example.org/')
    term = EX.a

    assert isinstance(term, Term)
    assert term == 'http://example.org/a'


def test_uri_property_returns_term():
    EX = Namespace('http://example.org/')
    term = EX.uri

    assert isinstance(term, Term)
    assert term == 'http://example.org/'


def test_terminates_with_adds_delimiter():
    EX = Namespace('http://example.org')
    EX2 = EX.terminates_with('/')
    assert str(EX2) == 'http://example.org/'
    assert isinstance(EX2, Namespace)
    # Original namespace is unchanged
    assert str(EX) == 'http://example.org'


def test_terminates_with_no_double_delimiter():
    EX = Namespace('http://example.org/')
    EX2 = EX.terminates_with('/')
    assert str(EX2) == 'http://example.org/'
    # Should return self if already ends with delimiter
    assert EX2 is EX or str(EX2) == str(EX)


def test_terminates_with_hash_namespace_raises():
    BASE = Namespace('http://example.org#')
    with pytest.raises(ValueError, match='Cannot set a trailing delimiter on a hash namespace'):
        BASE.terminates_with('/')


def test_terminates_with_invalid_character():
    EX = Namespace('http://example.org')
    with pytest.raises(ValueError):
        EX.terminates_with('')


def test_terminates_with_allows_concatenation():
    EX = Namespace('http://example.org').terminates_with('/')
    term = EX / 'a' / 'b' + 'c'
    assert str(term) == 'http://example.org/a/b/c'


def test_terminates_with_does_not_affect_getattr():
    EX = Namespace('http://example.org').terminates_with('/')
    t = EX.a
    assert isinstance(t, Term)
    assert str(t) == 'http://example.org/a'


def test_terminates_with_returns_new_instance():
    EX = Namespace('http://example.org')
    EX2 = EX.terminates_with('/')
    EX3 = EX.terminates_with('/')
    # New instances may be returned even if logically equal
    assert str(EX2) == str(EX3)
    assert isinstance(EX2, Namespace)



class Term(str):
    """A terminal URI value"""
    __slots__ = ()

    def __new__(cls, value: str):
        return super().__new__(cls, value)

    def __repr__(self) -> str:
        return f'Term({super().__repr__()})'
